@@ -24,7 +24,15 @@ __all__ = [
 ]
 
 
-ORGANIZATION_REGEX = r"(([\da-zA-Z])([_\w-]{,62})\.){,127}(([\da-zA-Z])[_\w-]{,61})?([\da-zA-Z]\.((xn\-\-[a-zA-Z\d]+)|([a-zA-Z\d]{2,})))"
+ORGANIZATION_REGEX = r"(?!-)([A-Za-z0-9-.]{5,50})"
+# Organization regex constraint to enforce that:
+# * Cannot contain a forward slash (/)
+# * Cannot solely consist of a single period (.) or double periods (..)
+# * Cannot match the regular expression: __.*__
+# * Cannot start with dash (-)
+# * Must be between at least 5 characters long and no longer than 50
+# * Must match domain names but also allow non-domain names and names including no period (.)
+
 NAME_REGEX = r"[a-zA-Z\_\-\.0-9]{1,64}"
 OPTIMIZER_ID_REGEX = f"^{ORGANIZATION_REGEX}/{NAME_REGEX}$"
 
@@ -47,7 +55,7 @@ class Optimizer(pydantic.BaseSettings):
 
     id: pydantic.constr(regex=OPTIMIZER_ID_REGEX)
     token: pydantic.SecretStr
-    base_url: pydantic.AnyHttpUrl = "https://api.opsani.com/"
+    base_url: pydantic.AnyHttpUrl = "https://api.opsani.com"
     _organization: str
     _name: str
     __url__: Optional[pydantic.AnyHttpUrl] = None
@@ -60,6 +68,10 @@ class Optimizer(pydantic.BaseSettings):
         self._name = name
         self.__url__ = __url__
 
+    @pydantic.validator("base_url")
+    def _rstrip_slash(cls, url: str) -> str:
+        return url.rstrip("/")
+
     @property
     def organization(self) -> str:
         """Returns the organization component of the optimizer ID.
@@ -91,7 +103,7 @@ class Optimizer(pydantic.BaseSettings):
         """
         return (
             self.__url__
-            or f"{self.base_url}accounts/{self.organization}/applications/{self.name}/"
+            or f"{self.base_url}/accounts/{self.organization}/applications/{self.name}/"
         )
 
     class Config:
@@ -99,6 +111,7 @@ class Optimizer(pydantic.BaseSettings):
         case_sensitive = True
         extra = pydantic.Extra.forbid
         underscore_attrs_are_private = True
+        validate_assignment = True
         fields = {
             "token": {
                 "env": "OPSANI_TOKEN",
