@@ -1,7 +1,8 @@
+import abc
 import json
 import operator
 import os
-from typing import List, Optional, Union
+from typing import Dict, List, Optional, Union, Type
 
 import kubernetes_asyncio
 import pydantic
@@ -19,7 +20,7 @@ KUBERNETES_PERMISSIONS = [
     servo.connectors.kubernetes.PermissionSet(
         group="",
         resources=["namespaces"],
-        verbs=["get", "list"],
+        verbs=["get"],
     ),
     servo.connectors.kubernetes.PermissionSet(
         group="",
@@ -48,7 +49,8 @@ class Memory(servo.connectors.kubernetes.Memory):
 
 class OpsaniDevConfiguration(servo.BaseConfiguration):
     namespace: str
-    deployment: str
+    deployment: Optional[str]
+    rollout: Optional[str]
     container: str
     service: str
     port: Optional[Union[pydantic.StrictInt, str]] = None
@@ -60,6 +62,16 @@ class OpsaniDevConfiguration(servo.BaseConfiguration):
         description="Duration to observe the application after an adjust to ensure the deployment is stable. May be overridden by optimizer supplied `control.adjust.settlement` value."
     )
 
+    @pydantic.root_validator
+    def check_deployment_and_rollout(cls, values):
+        if values.get('deployment') is not None and values.get('rollout') is not None:
+            raise ValueError("Configuration cannot specify both rollout and deployment")
+
+        if values.get('deployment') is None and values.get('rollout') is None:
+            raise ValueError("Configuration must specify either rollout or deployment")
+
+        return values
+
     @classmethod
     def generate(cls, **kwargs) -> "OpsaniDevConfiguration":
         return cls(
@@ -79,32 +91,38 @@ class OpsaniDevConfiguration(servo.BaseConfiguration):
         Returns:
             A Kubernetes connector configuration object.
         """
+        main_config = servo.connectors.kubernetes.DeploymentConfiguration(
+            name=(self.deployment or self.rollout),
+            strategy=servo.connectors.kubernetes.CanaryOptimizationStrategyConfiguration(
+                type=servo.connectors.kubernetes.OptimizationStrategy.canary,
+                alias="tuning"
+            ),
+            replicas=servo.Replicas(
+                min=0,
+                max=1,
+            ),
+            containers=[
+                servo.connectors.kubernetes.ContainerConfiguration(
+                    name=self.container,
+                    alias="main",
+                    cpu=self.cpu,
+                    memory=self.memory,
+                )
+            ],
+        )
+        if self.deployment:
+            main_arg = { 'deployments': [ main_config ] }
+        elif self.rollout:
+            main_arg = { 'rollouts': [ servo.connectors.kubernetes.RolloutConfiguration.parse_obj(
+                main_config.dict(exclude_none=True)
+            ) ] }
+
         return servo.connectors.kubernetes.KubernetesConfiguration(
             namespace=self.namespace,
             description="Update the namespace, deployment, etc. to match your Kubernetes cluster",
             timeout=self.timeout,
             settlement=self.settlement,
-            deployments=[
-                servo.connectors.kubernetes.DeploymentConfiguration(
-                    name=self.deployment,
-                    strategy=servo.connectors.kubernetes.CanaryOptimizationStrategyConfiguration(
-                        type=servo.connectors.kubernetes.OptimizationStrategy.canary,
-                        alias="tuning"
-                    ),
-                    replicas=servo.Replicas(
-                        min=0,
-                        max=1,
-                    ),
-                    containers=[
-                        servo.connectors.kubernetes.ContainerConfiguration(
-                            name=self.container,
-                            alias="main",
-                            cpu=self.cpu,
-                            memory=self.memory,
-                        )
-                    ],
-                )
-            ],
+            **main_arg,
             **kwargs,
         )
 
@@ -132,59 +150,44 @@ class OpsaniDevConfiguration(servo.BaseConfiguration):
                     servo.types.Unit.count,
                     query='envoy_cluster_membership_healthy{opsani_role="tuning"}',
                 ),
-                servo.connectors.prometheus.PrometheusMetric(
-                    "main_pod_avg_request_rate",
-                    servo.types.Unit.requests_per_second,
-                    query='avg(rate(envoy_cluster_upstream_rq_total{opsani_role!="tuning"}[1m]))',
-                ),
                 servo.connectors.prometheus.PrometheusMetric(
                     "total_request_rate",
                     servo.types.Unit.requests_per_second,
                     query="sum(rate(envoy_cluster_upstream_rq_total[1m]))",
                 ),
                 servo.connectors.prometheus.PrometheusMetric(
-                    "main_request_rate",
+                    "main_request_rate_avg",
                     servo.types.Unit.requests_per_second,
-                    query='sum(rate(envoy_cluster_upstream_rq_total{opsani_role!="tuning"}[1m]))',
+                    query='avg(rate(envoy_cluster_upstream_rq_total{opsani_role!="tuning"}[1m]))',
                 ),
                 servo.connectors.prometheus.PrometheusMetric(
-                    "tuning_request_rate",
+                    "tuning_request_rate_avg",
                     servo.types.Unit.requests_per_second,
-                    query='rate(envoy_cluster_upstream_rq_total{opsani_role="tuning"}[1m])',
+                    query='avg(rate(envoy_cluster_upstream_rq_total{opsani_role="tuning"}[1m]))',
                 ),
                 servo.connectors.prometheus.PrometheusMetric(
-                    "main_success_rate",
+                    "main_success_rate_avg",
                     servo.types.Unit.requests_per_second,
-                    query='sum(rate(envoy_cluster_upstream_rq_xx{opsani_role!="tuning", envoy_response_code_class=~"2|3"}[1m]))',
+                    query='avg(rate(envoy_cluster_upstream_rq_xx{opsani_role!="tuning", envoy_response_code_class=~"2|3"}[1m]))',
                 ),
                 servo.connectors.prometheus.PrometheusMetric(
-                    "tuning_success_rate",
+                    "tuning_success_rate_avg",
                     servo.types.Unit.requests_per_second,
-                    query='sum(rate(envoy_cluster_upstream_rq_xx{opsani_role="tuning", envoy_response_code_class=~"2|3"}[1m]))',
+                    query='avg(rate(envoy_cluster_upstream_rq_xx{opsani_role="tuning", envoy_response_code_class=~"2|3"}[1m]))',
                     absent=servo.connectors.prometheus.AbsentMetricPolicy.zero
                 ),
                 servo.connectors.prometheus.PrometheusMetric(
-                    "main_error_rate",
+                    "main_error_rate_avg",
                     servo.types.Unit.requests_per_second,
-                    query='sum(rate(envoy_cluster_upstream_rq_xx{opsani_role!="tuning", envoy_response_code_class=~"4|5"}[1m]))',
+                    query='avg(rate(envoy_cluster_upstream_rq_xx{opsani_role!="tuning", envoy_response_code_class=~"4|5"}[1m]))',
                     absent=servo.connectors.prometheus.AbsentMetricPolicy.zero
                 ),
                 servo.connectors.prometheus.PrometheusMetric(
-                    "tuning_error_rate",
+                    "tuning_error_rate_avg",
                     servo.types.Unit.requests_per_second,
-                    query='sum(rate(envoy_cluster_upstream_rq_xx{opsani_role="tuning", envoy_response_code_class=~"4|5"}[1m]))',
+                    query='avg(rate(envoy_cluster_upstream_rq_xx{opsani_role="tuning", envoy_response_code_class=~"4|5"}[1m]))',
                     absent=servo.connectors.prometheus.AbsentMetricPolicy.zero
                 ),
-                servo.connectors.prometheus.PrometheusMetric(
-                    "main_p99_latency",
-                    servo.types.Unit.milliseconds,
-                    query='avg(histogram_quantile(0.99,rate(envoy_cluster_upstream_rq_time_bucket{opsani_role!="tuning"}[1m])))',
-                ),
-                servo.connectors.prometheus.PrometheusMetric(
-                    "tuning_p99_latency",
-                    servo.types.Unit.milliseconds,
-                    query='avg(histogram_quantile(0.99,rate(envoy_cluster_upstream_rq_time_bucket{opsani_role="tuning"}[1m])))',
-                ),
                 servo.connectors.prometheus.PrometheusMetric(
                     "main_p90_latency",
                     servo.types.Unit.milliseconds,
@@ -209,10 +212,50 @@ class OpsaniDevConfiguration(servo.BaseConfiguration):
             **kwargs,
         )
 
-
-class OpsaniDevChecks(servo.BaseChecks):
+class BaseOpsaniDevChecks(servo.BaseChecks, abc.ABC):
     config: OpsaniDevConfiguration
 
+    @property
+    @abc.abstractmethod
+    def controller_type_name(self) -> str:
+        ...
+
+    @property
+    @abc.abstractmethod
+    def config_controller_name(self) -> str:
+        ...
+
+    @property
+    @abc.abstractmethod
+    def controller_class(self) -> Type[Union[
+        servo.connectors.kubernetes.Deployment,
+        servo.connectors.kubernetes.Rollout
+    ]]:
+        ...
+
+    @property
+    @abc.abstractmethod
+    def required_permissions(self) -> List[servo.connectors.kubernetes.PermissionSet]:
+        ...
+
+    @abc.abstractmethod
+    async def _get_port_forward_target(self) -> str:
+        ...
+
+    @abc.abstractmethod
+    def _get_generated_controller_config(self, config: servo.connectors.kubernetes.KubernetesConfiguration) -> Union[
+        servo.connectors.kubernetes.DeploymentConfiguration,
+        servo.connectors.kubernetes.RolloutConfiguration
+    ]:
+        ...
+
+    @abc.abstractmethod
+    def _get_controller_service_selector(self, controller: Union[
+        servo.connectors.kubernetes.Deployment,
+        servo.connectors.kubernetes.Rollout
+    ]) -> Dict[str, str]:
+        ...
+
     ##
     # Kubernetes essentials
 
@@ -235,7 +278,7 @@ class OpsaniDevChecks(servo.BaseChecks):
     async def check_permissions(self) -> None:
         async with kubernetes_asyncio.client.api_client.ApiClient() as api:
             v1 = kubernetes_asyncio.client.AuthorizationV1Api(api)
-            for permission in KUBERNETES_PERMISSIONS:
+            for permission in self.required_permissions:
                 for resource in permission.resources:
                     for verb in permission.verbs:
                         attributes = kubernetes_asyncio.client.models.V1ResourceAttributes(
@@ -260,50 +303,53 @@ class OpsaniDevChecks(servo.BaseChecks):
     async def check_kubernetes_namespace(self) -> None:
         await servo.connectors.kubernetes.Namespace.read(self.config.namespace)
 
-    @servo.checks.require('Deployment "{self.config.deployment}" is readable')
-    async def check_kubernetes_deployment(self) -> None:
-        await servo.connectors.kubernetes.Deployment.read(self.config.deployment, self.config.namespace)
+    @servo.checks.require('{self.controller_type_name} "{self.config_controller_name}" is readable')
+    async def check_kubernetes_controller(self) -> None:
+        await self.controller_class.read(self.config_controller_name, self.config.namespace)
 
     @servo.checks.require('Container "{self.config.container}" is readable')
     async def check_kubernetes_container(self) -> None:
-        deployment = await servo.connectors.kubernetes.Deployment.read(
-            self.config.deployment, self.config.namespace
-        )
-        container = deployment.find_container(self.config.container)
+        controller = await self.controller_class.read(self.config_controller_name, self.config.namespace)
+        container = controller.find_container(self.config.container)
         assert (
             container
-        ), f"failed reading Container '{self.config.container}' in Deployment '{self.config.deployment}'"
+        ), f"failed reading Container '{self.config.container}' in {self.controller_type_name} '{self.config_controller_name}'"
 
     @servo.require('Container "{self.config.container}" has resource requirements')
     async def check_resource_requirements(self) -> None:
-        deployment = await servo.connectors.kubernetes.Deployment.read(self.config.deployment, self.config.namespace)
-        container = deployment.find_container(self.config.container)
+        controller = await self.controller_class.read(self.config_controller_name, self.config.namespace)
+        container = controller.find_container(self.config.container)
         assert container
         assert container.resources, "missing container resources"
 
         # Apply any defaults/overrides for requests/limits from config
         servo.connectors.kubernetes.set_container_resource_defaults_from_config(container, self.config)
 
-        # TODO: How do we handle when you just don't have any requests?
-        # assert container.resources.requests, "missing requests for container resources"
-        # assert container.resources.requests.get("cpu"), "missing request for resource 'cpu'"
-        # assert container.resources.requests.get("memory"), "missing request for resource 'memory'"
+        for resource in servo.connectors.kubernetes.Resource.values():
+            current_state = None
+            container_requirements = container.get_resource_requirements(resource)
+            get_requirements = getattr(self.config, resource).get
+            for requirement in get_requirements:
+                current_state = container_requirements.get(requirement)
+                if current_state:
+                    break
 
-        # assert container.resources.limits, "missing limits for container resources"
-        # assert container.resources.limits.get("cpu"), "missing limit for resource 'cpu'"
-        # assert container.resources.limits.get("memory"), "missing limit for resource 'memory'"
+            assert current_state, (
+                f"{self.controller_type_name} {self.config_controller_name} target container {self.config.container} spec does not define the resource {resource}. "
+                f"At least one of the following must be specified: {', '.join(map(lambda req: req.resources_key, get_requirements))}"
+            )
 
     @servo.checks.require("Target container resources fall within optimization range")
     async def check_target_container_resources_within_limits(self) -> None:
-        # Load the Deployment
-        deployment = await servo.connectors.kubernetes.Deployment.read(
-            self.config.deployment,
+        # Load the Controller
+        controller = await self.controller_class.read(
+            self.config_controller_name,
             self.config.namespace
         )
-        assert deployment, f"failed to read deployment '{self.config.deployment}' in namespace '{self.config.namespace}'"
+        assert controller, f"failed to read {self.controller_type_name} '{self.config_controller_name}' in namespace '{self.config.namespace}'"
 
         # Find the target Container
-        target_container = next(filter(lambda c: c.name == self.config.container, deployment.containers), None)
+        target_container = next(filter(lambda c: c.name == self.config.container, controller.containers), None)
         assert target_container, f"failed to find container '{self.config.container}' when verifying resource limits"
 
         # Apply any defaults/overrides from the config
@@ -335,11 +381,11 @@ class OpsaniDevChecks(servo.BaseChecks):
         assert container_memory_value >= config_memory_min, f"target container Memory value {container_memory_value.human_readable()} must be greater than optimizable minimum {config_memory_min.human_readable()}"
         assert container_memory_value <= config_memory_max, f"target container Memory value {container_memory_value.human_readable()} must be less than optimizable maximum {config_memory_max.human_readable()}"
 
-    @servo.require('Deployment "{self.config.deployment}" is ready')
-    async def check_deployment(self) -> None:
-        deployment = await servo.connectors.kubernetes.Deployment.read(self.config.deployment, self.config.namespace)
-        if not await deployment.is_ready():
-            raise RuntimeError(f'Deployment "{deployment.name}" is not ready')
+    @servo.require('{self.controller_type_name} "{self.config_controller_name}"  is ready')
+    async def check_controller_readiness(self) -> None:
+        controller = await self.controller_class.read(self.config_controller_name, self.config.namespace)
+        if not await controller.is_ready():
+            raise RuntimeError(f'{self.controller_type_name} "{controller.name}" is not ready')
 
     @servo.checks.require("service")
     async def check_kubernetes_service(self) -> None:
@@ -384,21 +430,19 @@ class OpsaniDevChecks(servo.BaseChecks):
 
         return f"Service Port: {port.name} {port.port}:{port.target_port}/{port.protocol}"
 
-    @servo.checks.check('Service routes traffic to Deployment Pods')
-    async def check_service_routes_traffic_to_deployment(self) -> None:
+    @servo.checks.check('Service routes traffic to {self.controller_type_name} Pods')
+    async def check_service_routes_traffic_to_controller(self) -> None:
         service = await servo.connectors.kubernetes.Service.read(
             self.config.service, self.config.namespace
         )
-        deployment = await servo.connectors.kubernetes.Deployment.read(
-            self.config.deployment, self.config.namespace
-        )
+        controller = await self.controller_class.read(self.config_controller_name, self.config.namespace)
 
-        # NOTE: The Service labels should be a subset of the Deployment labels
-        deployment_labels = deployment.obj.spec.selector.match_labels
-        delta = dict(set(service.selector.items()) - set(deployment_labels.items()))
+        # NOTE: The Service labels should be a subset of the controller labels
+        controller_labels = self._get_controller_service_selector(controller)
+        delta = dict(set(service.selector.items()) - set(controller_labels.items()))
         if delta:
             desc = ' '.join(map('='.join, delta.items()))
-            raise RuntimeError(f"Service selector does not match Deployment labels. Missing labels: {desc}")
+            raise RuntimeError(f"Service selector does not match {self.controller_type_name} labels. Missing labels: {desc}")
 
     ##
     # Prometheus sidecar
@@ -529,47 +573,41 @@ class OpsaniDevChecks(servo.BaseChecks):
         return pods
 
     ##
-    # Kubernetes Deployment edits
+    # Kubernetes Controller edits
 
-    @servo.checks.require("Deployment PodSpec has expected annotations")
-    async def check_deployment_annotations(self) -> None:
-        deployment = await servo.connectors.kubernetes.Deployment.read(
-            self.config.deployment,
-            self.config.namespace
-        )
-        assert deployment, f"failed to read deployment '{self.config.deployment}' in namespace '{self.config.namespace}'"
+    @servo.checks.require("{self.controller_type_name} PodSpec has expected annotations")
+    async def check_controller_annotations(self) -> None:
+        controller = await self.controller_class.read(self.config_controller_name, self.config.namespace)
+        assert controller, f"failed to read {self.controller_type_name} '{self.config_controller_name}' in namespace '{self.config.namespace}'"
 
         # Add optimizer annotation to the static Prometheus values
         required_annotations = PROMETHEUS_ANNOTATION_DEFAULTS.copy()
         required_annotations['servo.opsani.com/optimizer'] = self.config.optimizer.id
 
         # NOTE: Only check for annotation keys
-        annotations = deployment.pod_template_spec.metadata.annotations or dict()
+        annotations = controller.pod_template_spec.metadata.annotations or dict()
         actual_annotations = set(annotations.keys())
         delta = set(required_annotations.keys()).difference(actual_annotations)
         if delta:
             annotations = dict(map(lambda k: (k, required_annotations[k]), delta))
             patch = {"spec": {"template": {"metadata": {"annotations": annotations}}}}
             patch_json = json.dumps(patch, indent=None)
-            command = f"kubectl --namespace {self.config.namespace} patch deployment {self.config.deployment} -p '{patch_json}'"
+            # NOTE: custom resources don't support strategic merge type. json merge is acceptable for both cases because the patch json doesn't contain lists
+            command = f"kubectl --namespace {self.config.namespace} patch {self.controller_type_name} {self.config_controller_name} --type='merge' -p '{patch_json}'"
             desc = ', '.join(sorted(delta))
             raise servo.checks.CheckError(
-                f"deployment '{deployment.name}' is missing annotations: {desc}",
+                f"{self.controller_type_name} '{controller.name}' is missing annotations: {desc}",
                 hint=f"Patch annotations via: `{command}`",
                 remedy=lambda: _stream_remedy_command(command)
             )
 
-    @servo.checks.require("Deployment PodSpec has expected labels")
-    async def check_deployment_labels(self) -> None:
-        deployment = await servo.connectors.kubernetes.Deployment.read(
-            self.config.deployment,
-            self.config.namespace
-        )
-        assert deployment, f"failed to read deployment '{self.config.deployment}' in namespace '{self.config.namespace}'"
-
-        labels = deployment.pod_template_spec.metadata.labels
-        assert labels, f"deployment '{deployment.name}' does not have any labels"
+    @servo.checks.require("{self.controller_type_name} PodSpec has expected labels")
+    async def check_controller_labels(self) -> None:
+        controller = await self.controller_class.read(self.config_controller_name, self.config.namespace)
+        assert controller, f"failed to read {self.controller_type_name} '{self.config_controller_name}' in namespace '{self.config.namespace}'"
 
+        labels = controller.pod_template_spec.metadata.labels
+        assert labels, f"{self.controller_type_name} '{controller.name}' does not have any labels"
         # Add optimizer label to the static values
         required_labels = ENVOY_SIDECAR_LABELS.copy()
         required_labels['servo.opsani.com/optimizer'] = servo.connectors.kubernetes.dns_labelize(self.config.optimizer.id)
@@ -580,23 +618,21 @@ class OpsaniDevChecks(servo.BaseChecks):
             desc = ', '.join(sorted(map('='.join, delta.items())))
             patch = {"spec": {"template": {"metadata": {"labels": delta}}}}
             patch_json = json.dumps(patch, indent=None)
-            command = f"kubectl --namespace {self.config.namespace} patch deployment {self.config.deployment} -p '{patch_json}'"
+            # NOTE: custom resources don't support strategic merge type. json merge is acceptable for both cases because the patch json doesn't contain lists
+            command = f"kubectl --namespace {self.config.namespace} patch {self.controller_type_name} {controller.name} --type='merge' -p '{patch_json}'"
             raise servo.checks.CheckError(
-                f"deployment '{deployment.name}' is missing labels: {desc}",
+                f"{self.controller_type_name} '{controller.name}' is missing labels: {desc}",
                 hint=f"Patch labels via: `{command}`",
                 remedy=lambda: _stream_remedy_command(command)
             )
 
-    @servo.checks.require("Deployment has Envoy sidecar container")
-    async def check_deployment_envoy_sidecars(self) -> None:
-        deployment = await servo.connectors.kubernetes.Deployment.read(
-            self.config.deployment,
-            self.config.namespace
-        )
-        assert deployment, f"failed to read deployment '{self.config.deployment}' in namespace '{self.config.namespace}'"
+    @servo.checks.require("{self.controller_type_name} has Envoy sidecar container")
+    async def check_controller_envoy_sidecars(self) -> None:
+        controller = await self.controller_class.read(self.config_controller_name, self.config.namespace)
+        assert controller, f"failed to read {self.controller_type_name} '{self.config_controller_name}' in namespace '{self.config.namespace}'"
 
         # Search the containers list for the sidecar
-        for container in deployment.containers:
+        for container in controller.containers:
             if container.name == "opsani-envoy":
                 return
 
@@ -604,23 +640,24 @@ class OpsaniDevChecks(servo.BaseChecks):
             f" --port {self.config.port}" if self.config.port is not None
             else ''
         )
-        command = f"kubectl exec -n {self.config.namespace} -c servo {self._servo_resource_target} -- servo --token-file /servo/opsani.token inject-sidecar --namespace {self.config.namespace} --service {self.config.service}{port_switch} deployment/{self.config.deployment}"
+        command = (
+            f"kubectl exec -n {self.config.namespace} -c servo {self._servo_resource_target} -- "
+            f"servo --token-file /servo/opsani.token inject-sidecar --namespace {self.config.namespace} --service {self.config.service}{port_switch} "
+            f"{self.controller_type_name.lower()}/{self.config_controller_name}"
+        )
         raise servo.checks.CheckError(
-            f"deployment '{deployment.name}' pod template spec does not include envoy sidecar container ('opsani-envoy')",
+            f"{self.controller_type_name} '{controller.name}' pod template spec does not include envoy sidecar container ('opsani-envoy')",
             hint=f"Inject Envoy sidecar container via: `{command}`",
             remedy=lambda: _stream_remedy_command(command)
         )
 
     @servo.checks.require("Pods have Envoy sidecar containers")
     async def check_pod_envoy_sidecars(self) -> None:
-        deployment = await servo.connectors.kubernetes.Deployment.read(
-            self.config.deployment,
-            self.config.namespace
-        )
-        assert deployment, f"failed to read deployment '{self.config.deployment}' in namespace '{self.config.namespace}'"
+        controller = await self.controller_class.read(self.config_controller_name, self.config.namespace)
+        assert controller, f"failed to read {self.controller_type_name} '{self.config_controller_name}' in namespace '{self.config.namespace}'"
 
         pods_without_sidecars = []
-        for pod in await deployment.get_pods():
+        for pod in await controller.get_pods():
             # Search the containers list for the sidecar
             if not pod.get_container('opsani-envoy'):
                 pods_without_sidecars.append(pod)
@@ -664,9 +701,14 @@ class OpsaniDevChecks(servo.BaseChecks):
         assert response.data.result_type == servo.connectors.prometheus.ResultType.vector, f"expected a vector result but found {response.data.result_type}"
         assert len(response.data) == 1, f"expected Prometheus API to return a single result for metric '{metric.name}' but found {len(response.data)}"
         result = response.data[0]
-        timestamp, value = result.value
+        _, value = result.value
         if value in {None, 0.0}:
-            command = f"kubectl exec -n {self.config.namespace} -c servo {self._servo_resource_target} -- sh -c \"kubectl port-forward --namespace={self.config.namespace} deploy/{self.config.deployment} 9980 & echo 'GET http://localhost:9980/' | vegeta attack -duration 10s | vegeta report -every 3s\""
+            port_forward_target = await self._get_port_forward_target()
+            command = (
+                f"kubectl exec -n {self.config.namespace} -c servo {self._servo_resource_target} -- "
+                f"sh -c \"kubectl port-forward --namespace={self.config.namespace} {port_forward_target} 9980 & "
+                f"echo 'GET http://localhost:9980/' | vegeta attack -duration 10s | vegeta report -every 3s\""
+            )
             raise servo.checks.CheckError(
                 f"Envoy is not reporting any traffic to Prometheus for metric '{metric.name}' ({metric.query})",
                 hint=f"Send traffic to your application on port 9980. Try `{command}`",
@@ -701,9 +743,9 @@ class OpsaniDevChecks(servo.BaseChecks):
     async def check_tuning_is_running(self) -> None:
         # Generate a KubernetesConfiguration to initialize the optimization class
         kubernetes_config = self.config.generate_kubernetes_config()
-        deployment_config = kubernetes_config.deployments[0]
+        controller_config = self._get_generated_controller_config(kubernetes_config)
         optimization = await servo.connectors.kubernetes.CanaryOptimization.create(
-            deployment_config, timeout=kubernetes_config.timeout
+            controller_config, timeout=kubernetes_config.timeout
         )
 
         # Ensure the tuning pod is available
@@ -756,6 +798,121 @@ class OpsaniDevChecks(servo.BaseChecks):
         else:
             return "deployment/servo"
 
+class OpsaniDevChecks(BaseOpsaniDevChecks):
+    """Opsani dev checks against standard kubernetes Deployments"""
+
+    @property
+    def controller_type_name(self) -> str:
+        return "Deployment"
+
+    @property
+    def config_controller_name(self) -> str:
+        return self.config.deployment
+
+    @property
+    def controller_class(self) -> Type[servo.connectors.kubernetes.Deployment]:
+        return servo.connectors.kubernetes.Deployment
+
+    @property
+    def required_permissions(self) -> List[servo.connectors.kubernetes.PermissionSet]:
+        return KUBERNETES_PERMISSIONS
+
+    async def _get_port_forward_target(self) -> str:
+        return f"deploy/{self.config.deployment}"
+
+    def _get_generated_controller_config(
+        self,
+        config: servo.connectors.kubernetes.KubernetesConfiguration
+    ) -> servo.connectors.kubernetes.DeploymentConfiguration:
+        return config.deployments[0]
+
+    def _get_controller_service_selector(self, controller: servo.connectors.kubernetes.Deployment) -> Dict[str, str]:
+        return controller.obj.spec.selector.match_labels
+
+class OpsaniDevRolloutChecks(BaseOpsaniDevChecks):
+    """Opsani dev checks against argoproj.io Rollouts"""
+    @property
+    def controller_type_name(self) -> str:
+        return "Rollout"
+
+    @property
+    def config_controller_name(self) -> str:
+        return self.config.rollout
+
+    @property
+    def controller_class(self) -> Type[servo.connectors.kubernetes.Rollout]:
+        return servo.connectors.kubernetes.Rollout
+
+    @property
+    def required_permissions(self) -> List[servo.connectors.kubernetes.PermissionSet]:
+        return KUBERNETES_PERMISSIONS + [servo.connectors.kubernetes.PermissionSet(
+            group="argoproj.io",
+            resources=["rollouts", "rollouts/status"],
+            verbs=["get", "list", "watch", "update", "patch"],
+        )]
+
+    async def _get_port_forward_target(self) -> str:
+        # NOTE rollouts don't support kubectl port-forward, have to target the current replicaset instead
+        rollout = await servo.connectors.kubernetes.Rollout.read(
+            self.config.rollout,
+            self.config.namespace
+        )
+        assert rollout, f"failed to read rollout '{self.config.rollout}' in namespace '{self.config.namespace}'"
+        assert rollout.status, f"unable to verify envoy proxy. rollout '{self.config.rollout}' in namespace '{self.config.namespace}' has no status"
+        assert rollout.status.current_pod_hash, f"unable to verify envoy proxy. rollout '{self.config.rollout}' in namespace '{self.config.namespace}' has no currentPodHash"
+        return f"replicaset/{rollout.name}-{rollout.status.current_pod_hash}"
+
+    def _get_generated_controller_config(
+        self,
+        config: servo.connectors.kubernetes.KubernetesConfiguration
+    ) -> servo.connectors.kubernetes.RolloutConfiguration:
+        return config.rollouts[0]
+
+    def _get_controller_service_selector(self, controller: servo.connectors.kubernetes.Rollout) -> Dict[str, str]:
+        match_labels = dict(controller.obj.spec.selector.match_labels)
+        assert controller.status, f"unable to determine service selector. rollout '{self.config.rollout}' in namespace '{self.config.namespace}' has no status"
+        assert controller.status.current_pod_hash, f"unable to determine service selector. rollout '{self.config.rollout}' in namespace '{self.config.namespace}' has no currentPodHash"
+        match_labels['rollouts-pod-template-hash'] = controller.status.current_pod_hash
+        return match_labels
+
+    @servo.checks.require("Rollout Selector and PodSpec has opsani_role label")
+    async def check_rollout_selector_labels(self) -> None:
+        if os.environ.get("POD_NAME") and os.environ.get("POD_NAMESPACE"):
+            return # Setting owner reference to servo should prevent tuning pod from being adopted by the rollout controller
+
+        rollout = await servo.connectors.kubernetes.Rollout.read(self.config.rollout, self.config.namespace)
+        assert rollout, f"failed to read Rollout '{self.config.rollout}' in namespace '{self.config.namespace}'"
+
+        spec_patch = {}
+        match_labels = rollout.obj.spec.selector.match_labels or dict()
+        opsani_role_selector = match_labels.get("opsani_role")
+        if opsani_role_selector is None or opsani_role_selector == "tuning":
+            opsani_role_selector = "mainline"
+            spec_patch["selector"] = {"matchLabels": {"opsani_role": opsani_role_selector}}
+
+        labels = rollout.pod_template_spec.metadata.labels or dict()
+        opsani_role_label = labels.get("opsani_role")
+        if opsani_role_label is None or opsani_role_label == "tuning" or opsani_role_label != opsani_role_selector:
+            spec_patch["template"] = {"metadata": {"labels": {"opsani_role": opsani_role_selector }}}
+
+        if spec_patch: # Check failed if spec needs patching
+            patch = {"spec": spec_patch}
+            patch_json = json.dumps(patch, indent=None)
+            # NOTE: custom resources don't support strategic merge type. json merge is acceptable because the patch json doesn't contain lists
+            command = f"kubectl --namespace {self.config.namespace} patch rollout {self.config.rollout} --type='merge' -p '{patch_json}'"
+            replicasets = [ f"rs/{rollout.name}-{rollout.status.current_pod_hash}" ]
+            if rollout.status.stable_RS and rollout.status.stable_RS != rollout.status.current_pod_hash:
+                replicasets.append(f"rs/{rollout.name}-{rollout.status.stable_RS}")
+            raise servo.checks.CheckError(
+                (
+                    f"Rollout '{self.config.rollout}' has missing/mismatched opsani_role selector and/or label."
+                    " Label opsani_role with value != \"tuning\" is required to prevent the rollout controller from adopting and destroying the tuning pod"
+                ),
+                hint=(
+                    f"NOTE: Running this patch will require that you manually scale down or delete the replicaset(s) ({', '.join(replicasets)})"
+                    f" orphaned by the selector update. Patch selector and labels via: `{command}`"
+                )
+            )
 
 @servo.metadata(
     description="Optimize a single service via a tuning instance and an Envoy sidecar",
@@ -791,7 +948,12 @@ class OpsaniDevConnector(servo.BaseConnector):
         matching: Optional[servo.CheckFilter],
         halt_on: Optional[servo.ErrorSeverity] = servo.ErrorSeverity.critical,
     ) -> List[servo.Check]:
-        return await OpsaniDevChecks.run(
+        if self.config.deployment:
+            checks_class = OpsaniDevChecks
+        elif self.config.rollout:
+            checks_class = OpsaniDevRolloutChecks
+
+        return await checks_class.run(
             self.config, matching=matching, halt_on=halt_on
         )
 
